/*----------------------------------------------------------------------
    NHocr - Japanese OCR     nhocr  ver 0.16
        Written by H.Goto, Jan. 2008
        Revised by H.Goto, Feb. 2008
        Revised by H.Goto, Sep. 2008
        Revised by H.Goto, Oct. 2008
        Revised by H.Goto, Dec. 2008
        Revised by H.Goto, May 15, 2009
----------------------------------------------------------------------*/

/*--------------
  Copyright 2008,2009  Hideaki Goto

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at
      http://www.apache.org/licenses/LICENSE-2.0
  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
--------------*/


#include        <stdio.h>
#include        <stdlib.h>
#include        <string.h>
#include        <math.h>

#include	"utypes.h"

#include	"ufilep.h"
#include	"siplib.h"
#include	"imgobj.h"
#include	"objgrp.h"

#include	"nhocr.h"


static int	sw_debug = 0;
static int	sw_verbose = 0;
static int	sw_line = 0;	// Line recognition mode
static int	sw_interpolate = 1;

#define	MaxLineLength	1024
char	resultline[MaxLineLength +1];

#define	debugprintf	if(sw_debug)printf


/*------------------------------------------------------
	Load / Save Image File
------------------------------------------------------*/

static int LoadImage(char *fname,SIPImage **image,int depthconv, \
		int fgcol,int bgcol){
	PBMFILE	PBMLD;
	int	retcode,depth,x,y;
	SIPImage	*imgbuf;
	uchar	*lbuf,*p;
	*image = 0;	/* default value = NULL */
	PBMLD.setpal_fb((uchar)fgcol,(uchar)bgcol);
	retcode = PBMLD.open(fname,"r");
	if ( retcode )  return(retcode);
	switch ( PBMLD.getfiletype() ){
	  case 0:	depth = 1;  break;
	  case 1:	depth = 8;  break;
	  case 2:	depth = 32;  break;
			/* 24bit in file, but 32bit on memory */
	  default:	return(-5);
	}
	if ( depthconv == 24 )  depthconv = 32;
	switch ( depthconv ){
	  case 0:	depthconv = depth;  break;
	  case 1:	if ( depth > 1 )  return(-5);
			break;
	  case 8:	if ( depth > 8 )  return(-5);
			break;
	  case 32:	break;
	  default:	return(-5);
	}
	if ( 0 == (imgbuf = sip_CreateImage(PBMLD.getwidth(), \
		PBMLD.getheight(),depthconv)) )  return(-6);
	if ( 0 == (lbuf = new uchar [3 * PBMLD.getwidth() ]) )  return(-6);
	for ( y=0 ; y < PBMLD.getheight() ; y++ ){
		if ( depthconv == depth && depth < 32 ){
			if ( 0 != PBMLD.readline(-1,\
				(void *)imgbuf->pdata[y]) ){
				delete []lbuf;
				return(-1);	/* read error */
			}
		}
		else if ( depthconv == depth && depth == 32 ){
			if ( 0 != PBMLD.readline(-1,(void *)lbuf) ){
				delete []lbuf;
				return(-1);	/* read error */
			}
			p = (uchar *)imgbuf->pdata[y];
			for ( x=0 ; x < PBMLD.getwidth() ; x++ ){
				p[0] = lbuf[3*x  ];	/* R */
				p[1] = lbuf[3*x+1];	/* G */
				p[2] = lbuf[3*x+2];	/* B */
				p += 4;
			}
		}
		else if ( depthconv < 32 ){
			if ( 0 != PBMLD.readline_gray(-1,\
				(void *)imgbuf->pdata[y]) ){
				delete []lbuf;
				return(-1);	/* read error */
			}
		}
		else{	if ( 0 != PBMLD.readline_gray(-1,(void *)lbuf) ){
				delete []lbuf;
				return(-1);	/* read error */
			}
			p = (uchar *)imgbuf->pdata[y];
			for ( x=0 ; x < PBMLD.getwidth() ; x++ ){
				p[0] = lbuf[x];	/* R=Y */
				p[1] = lbuf[x];	/* G=Y */
				p[2] = lbuf[x];	/* B=Y */
				p += 4;
			}
		}
	}
	delete []lbuf;
	*image = imgbuf;
	return(0);
}


static int SaveImage(char *fname,SIPImage *image){
	int	x,y,retcode;
	uchar	*lbuf,*p;
	int	depth;
	PBMFILE	PBMSV;
	if ( image->depth != 32 )  depth = image->depth;  else  depth = 24;
	if ( PBMSV.setsize(image->width,image->height,depth) ){
		return(-5);	/* illegal size */
	}
	if ( 0 == (lbuf = new uchar [3 * image->width]) ){
		return(-6);	/* memory error */
	}
	retcode = PBMSV.open(fname,"w");
	if ( 0 > retcode )  return(retcode);
	retcode = 0;
	if ( image->depth == 1 || image->depth == 8 ){
		for ( y=0 ; y < image->height ; y++ ){
			if ( PBMSV.writeline(-1,(void *)image->pdata[y]) ){
				retcode = -1;  break;
			}
		}
	}
	else{	for ( y=0 ; y < image->height ; y++ ){
			p = (uchar *)image->pdata[y];
			for ( x=0 ; x<image->width ; x++ ){
				lbuf[3*x  ] = p[0];
				lbuf[3*x+1] = p[1];
				lbuf[3*x+2] = p[2];
				p += 4;
			}
			if ( PBMSV.writeline(-1,(void *)lbuf) ){
				retcode = -1;
				break;
			}
		}
	}
	delete []lbuf;
	if ( retcode )  PBMSV.close();  else  retcode = PBMSV.close();
	return(retcode);
}




/*----------------------------------------------
    Load image file and convert it to grayscale
----------------------------------------------*/

static int load_image(char *fname,SIPImage **image,int *width,int *height){
	PBMFILE	PBMLD;
	int	retcode,depth,x,y;
	SIPImage	*imgbuf;
	uchar	*lbuf, *p;
	double	gray;
	int	bufwidth,bufheight;
	*image = 0;	/* default value = NULL */
	PBMLD.setpal_fb(0,0xff);
	retcode = PBMLD.open(fname,"r");
	if ( retcode )  return(retcode);
	switch ( PBMLD.getfiletype() ){
	  case 0:	depth = 1;  break;
	  case 1:	depth = 8;  break;
	  case 2:	depth = 24;  break;
	  default:	return(-5);
	}
	bufwidth  = *width  = PBMLD.getwidth();
	bufheight = *height = PBMLD.getheight();

	if ( 0 == (imgbuf = sip_CreateImage(bufwidth,bufheight,8)) )  return(-6);
	if ( 0 == (lbuf = new uchar[*width * 3]) ){
		sip_DestroyImage(imgbuf);
		return(-6);
	}
	sip_ClearImage(imgbuf,0xff);
	for ( y=0 ; y < *height ; y++ ){
		p = (uchar *)imgbuf->pdata[y];
		if ( 24 == depth ){
			if ( 0 != PBMLD.readline(-1,\
				(void *)lbuf) ){
				delete []lbuf;
				return(-1);	/* read error */
			}
			for ( x=0 ; x<*width ; x++ ){
				gray  = 0.299 * (double)lbuf[3*x];
				gray += 0.587 * (double)lbuf[3*x+1];
				gray += 0.114 * (double)lbuf[3*x+2];
				if ( gray>255 )  gray=255;
				p[x] = (uchar)((uint)gray);
			}
		}
		else if ( 8 == depth ){
			if ( 0 != PBMLD.readline(-1,(void *)p) ){
				delete []lbuf;
				return(-1);	/* read error */
			}
		}
		else{	if ( 0 != PBMLD.readline_gray(-1,(void *)p) ){
				delete []lbuf;
				return(-1);	/* read error */
			}
		}
		for ( x=*width ; x<bufwidth ; x++ )  p[x] = p[*width-1];
	}
	for ( ; y < bufheight ; y++ ){
		memcpy((void*)imgbuf->pdata[y], (void*)imgbuf->pdata[*height-1],bufwidth);
	}

	*image = imgbuf;
	PBMLD.close();
	delete []lbuf;
	return(0);
}




/*----------------------------------------------
    Convert gray image into binary and save it
----------------------------------------------*/

static int save_bin_image(char *fname, SIPImage *image){
	int	x,y;
	uchar	*p;
	SIPImage	*bimage;

	if ( 0 == (bimage = sip_CreateImage(image->width, image->height,1)) ){
		fputs("Low memory.\n",stderr);
		return(-1);
	}

	for ( y=0 ; y<image->height ; y++ ){
		sip_cvt8to1u((uchar *)image->pdata[y], \
			(uchar *)bimage->pdata[y],0,image->width,127);
		p = (uchar *)bimage->pdata[y];
		for ( x=0 ; x < (image->width +7) /8 ; x++ )  p[x] = ~p[x];
	}

	if ( 0 > SaveImage(fname,bimage) ){
		sip_DestroyImage(bimage);
		return(-1);
	}
	sip_DestroyImage(bimage);
	return(0);
}




/*----------------------------------------------
    Print usage
----------------------------------------------*/

static void usage(){
	fputs("NHocr - Japanese OCR  v0.16  Copyright (C) 2008,2009 Hideaki Goto \n",stderr);
	fputs("usage: nhocr [-options] [-o result.txt] img_file [img_file2 img_file3 ...]\n",stderr);
//	fputs("         -i        :  enable interpolation in adaptive mode \n",stderr);
	fputs("         -o -      :  write text data to stdout \n",stderr);
	fputs("         -line     :  line recognition mode \n",stderr);
	fputs("         -csize N  :  cell size in adaptive binarization (default: 16) \n",stderr);
//	fputs("           (smaller value leads to critical failure of the bi-modality check) \n",stderr);
//	fputs("         -verbose  :  show process information \n",stderr);
	fputs("\ndescription:\n",stderr);
	fputs("  The current version of nhocr does not support \"Page recognition\n  mode\". Option -line is always required.\n",stderr);
	fputs("  nhocr reads PBM/PGM/PPM image file(s), recognizes the text\n",stderr);
	fputs("  line image for each file, and produces text data in UTF-8.\n",stderr);
	fputs("  Each file should contain only ONE horizontal text line image\n",stderr);
	fputs("  without any surrounding dirt.\n",stderr);
}




/*----------------------------------------------
    Main routine
----------------------------------------------*/

int main(int ac,char *av[]){
	char	*infile  = NULL;
	char	*outfile = NULL;
	FILE	*fp_out = stdout;
	int	width=0,height=0;
	int	csize = 16;
	int	i,k;
	int	argn = 1;
	SIPImage	*image, *bimage;
	NHrec	NHrec;

	for ( k=1, i=1 ; i<ac && k ; i++ ){
		if ( 0 == strcmp("-csize",av[i]) ){
			if ( ++i >= ac ){  k=0; continue; }
			csize = atoi(av[i]);
			if ( csize < 4 ){  k=0; continue; }
			continue;
		}
		if ( 0 == strcmp("-o",av[i]) ){
			if ( ++i >= ac ){  k=0; continue; }
			outfile = av[i];
			continue;
		}
		if ( 0 == strcmp("-debug",av[i]) ){  sw_debug = -1;  continue; }
		if ( 0 == strcmp("-i",av[i]) ){  sw_interpolate = -1;  continue; }
		if ( 0 == strcmp("-page",av[i]) ){  sw_line = 0;  continue; }
		if ( 0 == strcmp("-line",av[i]) ){  sw_line = -1;  continue; }
		if ( 0 == strcmp("-verbose",av[i]) ){  sw_verbose = -1;  continue; }
		if ( 0 == strcmp("-v",av[i]) ){  sw_verbose = -1;  continue; }
		if ( 0 == strcmp("-h",av[i]) ){  usage(); return(0); }
		if ( av[i][0] == '-' && strlen(av[i]) > 1 ){  k=0;  continue; }
		if ( NULL == infile ){
			infile = av[ argn = i ];
			break;
		}
	}
	if ( k == 0 || infile == NULL ){  usage();  return(0); }
	if ( k == 0 || outfile == NULL ){
		fprintf(stderr,"Output file is not specified.\n");
		return(1);
	}

	if ( sw_line == 0 ){
		fputs("The current version of nhocr supports \"Line recognition mode\" only.\n", stderr);
		fputs("Please specify -line option.\n", stderr);
		return(1);
	}


/* ---- Initialize line recognizer ---- */

	if ( sw_debug )  NHrec.debug = 1;

	if ( getenv("NHOCR_DICDIR") ){
		NHrec.setlibdir(getenv("NHOCR_DICDIR"));
		debugprintf("NHOCR_DICDIR= %s\n",getenv("NHOCR_DICDIR"));
	}

	if ( NHrec.open() ){
		return(2);
	}

	if ( 0 != strcmp(outfile, "-") ){
		if ( NULL == (fp_out = fopen(outfile,"wb")) ){
			fprintf(stderr,"Can't write \"%s\".\n",outfile);
			return(2);
		}
	}




  for ( ; argn < ac ; argn++ ){
	infile = av[argn];
	if ( infile[0] == '-' )  continue;

/* ---- Read Image ---- */

	if ( 0 > load_image(infile,&image,&width,&height) ){
		fprintf(stderr,"Can't read \"%s\".\n",infile);
		return(2);
	}

	if ( sw_verbose ){
		fprintf(stderr,"Input image size = %d x %d\n",width,height);
	}

	width  = image->width;
	height = image->height;


/* ---- Binarize image ---- */

	if ( 0 == (bimage = sip_DuplicateImage(image)) ){
		fputs("Low memory.\n",stderr);
		return(1);
	}

	NHrec.binarize_otsu(bimage,bimage,csize,sw_interpolate);

	NHrec.auto_invert(bimage);


/* ---- Recognize text line ---- */

	NHrec.rec_line(bimage, resultline, MaxLineLength);

	if ( sw_debug )  SaveImage("out.pgm",image);


/* ---- Write recognition result ---- */

	strcat(resultline, "\n");
	if ( EOF == fputs(resultline,fp_out) ){
		fprintf(stderr,"Can't write \"%s\".\n",outfile);
		return(2);
	}


/* ---- Save the binary image (debug) ---- */

	if ( sw_debug ){
		if ( 0 > save_bin_image("out2.pbm",bimage) ){
			fprintf(stderr,"Can't write \"%s\".\n","out2.pbm");
		}
		return(2);
	}


	sip_DestroyImage(bimage);
	sip_DestroyImage(image);
  }


	if ( fp_out != stdout ){
		if ( 0 != fclose(fp_out) ){
			fprintf(stderr,"Can't write \"%s\".\n",outfile);
		}
	}


/* ---- Close line recognizer ---- */

	NHrec.close();

	return(0);
}


